/**
* \file nonvolatile.c
* \brief  used for the non-volatile for NRF51.
*/


#include "../../driverInterfaces/nonvolatile.h"
#include "../../driverInterfaces/bleInterface.h"
#include "config.h"
#include "string.h"

/**
 * MACROS
 */



#define WORD_ALIGN(VALUE)	\
	if((uint32_t)VALUE & 0x3){ \
		VALUE += 4 - ((uint32_t)VALUE & 0x3);\
	}\
 
/**
 * ENUMS
 */

typedef enum
{
	BLOCK_WORKOUT_VARS = 0,
	BLOCK_CONFIG_VARS = 1,
	BLOCK_ERASED = 0xFFFFFFFF
} MEMORY_BLOCK;

typedef enum
{
	FLASH_BLANK_PAGE_ID		= 0xFF,
	FLASH_FILE_PAGE_ID 		= 1,
	FLASH_LOG_PAGE_ID		= 2,
	FLASH_INVALID_PAGE_ID   = 0 ///this is a bitwise OR of FILE PAGE and LOG PAGE
} FLASH_PAGE_ID;


typedef enum
{
	NO_UNCLEAN_PAGE = 0,
	PAGE_ONE_UNCLEAN = 1,
	PAGE_TWO_UNCLEAN = 2,
	PAGE_THREE_UNCLEAN = 3
} UNCLEAN_PAGE_NUMBERS;



/**
 * PRIVATE VARIABLES
 */
static NONVOLATILE_OBJ *nv;

//this is to simulate the logging section in the system
static volatile const uint8_t  __attribute__((used)) __attribute__((section(".nvdata.filedata")))
fileData[FILE_DATA_SIZE];


/**
 * PRIVATE FUNCTIONS
 */
/**
 * @fn flash_page_erase
 * @brief the lowest level function call that will erase a flash page
 * @param page_address a pointer to the address of the start of the page
 */
static void flash_page_erase(const volatile void *page_address);
/**
 * @fn flash_write
 * @brief the lowest level function call that writes to a flash address
 * @param address the address in NV memory to write to
 * @param value the pointer to the data buffer to write from
 * @param length the size in bytes to write.
 */
static void flash_write(const void *address, const void *value, uint16_t length);
/**
 * @fn readFileHeaders
 * @brief will read the fileData array to find which pages are file pages, and which are log pages
 */
static void readFileHeaders();

/**
 * @fn readDataSection
 * @brief will read the Data Page, clean an unclean page, and find the pointers to the log data
 * @return TRUE if dirty byte, False if clean byte
 */
static BOOL readDataSection();

/**
 * @fn readFileSection
 * @brief will read the File Page, clean an unclean page, and find the pointers to the files
 */
static void readFileSection();

/**
 * @fn fixUncleanPage
 * @brief the magical function that cleans our nonvolatile data after a crash, or after changing pages
 * @return TRUE if there was an unclean page and it was fixed, FALSE if there was no unclean page
 */
static BOOL fixUncleanPage();


/**
 * FUNCTION IMPLEMENTATIONS
 */

void NO_OPTIMIZATION initNonvolatile(TASK *nonvolatileTask, const uint8_t *configAddress)
{
	uint8_t memory_section_count;
	uint16_t data_block_size;
	uint16_t configSize;
	const uint8_t *memory_chunk_ids;
	const uint8_t 		*memory_chunk_sizes;
	const uint8_t			*memory_chunk_max_count;
	uint8_t 		i;
	uint16_t		currentMemoryReader;
	uint16_t currentMemoryLocation = sizeof(
										 NV_PAGE_HEADER); ///this number is the start of the actual memory portion of the file page

	///copy the configFile
	currentMemoryReader = 0;
	memcpy(&configSize, configAddress + currentMemoryReader, 2);
	currentMemoryReader = 2;

	memcpy(&data_block_size, configAddress + currentMemoryReader, 2);
	currentMemoryReader += 2;

	memcpy(&memory_section_count, configAddress + currentMemoryReader, 1);
	currentMemoryReader += 1;

	memory_chunk_ids = (uint8_t *)configAddress + currentMemoryReader;
	currentMemoryReader += memory_section_count * sizeof(uint8_t); ///the size of the MACHINE_CONFIG_IDs

	memory_chunk_sizes = configAddress + currentMemoryReader;
	currentMemoryReader += memory_section_count * sizeof(uint8_t);

	memory_chunk_max_count = configAddress + currentMemoryReader;
	currentMemoryReader += memory_section_count * sizeof(uint8_t);

	if(currentMemoryReader > configSize)
	{
		///todo handle this error case
	}
	///the word allign function will add extra numbers to any number passed to it such that it will be alligned to every 4 bytes in memory
	WORD_ALIGN(currentMemoryLocation);
	///allocate the memory
	nv = (NONVOLATILE_OBJ *)salloc(sizeof(NONVOLATILE_OBJ));
	///set the task. this may need to be tested to ensure that we're actually writing data
	nv->task = nonvolatileTask;
	nv->task->taskObject = nv;
	///assign the section count from config
	nv->memory_section_count = memory_section_count;
	///initialise busy bit
	nv->busyBit = 0;

	///this function will read the headers of each page in flash
	readFileHeaders();

	///we need to ensure that each size is word alligned so that any multiple of it will be word alligned.
	WORD_ALIGN(data_block_size);

	///now we write this data into our memory.
	nv->logDataPageObj.dataBlockSize = data_block_size;
	///determine the starting location pointer
	nv->logDataPageObj.dataStart = nv->currentLogPageStart + currentMemoryLocation;

	///calculate how large the section will be
	nv->logDataPageObj.dataSectionCount = ((1024 - currentMemoryLocation) / (data_block_size +
										   (1 * sizeof(uint8_t)))); ///we divide our memory between the two sections.
	///todo ensure that these will always be the same
	nv->logDataPageObj.dirtySectionCount =  nv->logDataPageObj.dataSectionCount;
	nv->logDataPageObj.dirtyByteStart = nv->logDataPageObj.dataStart +
										(nv->logDataPageObj.dataSectionCount * data_block_size);

	///allign where the dirtybyte will start(shouldn't change at all
	WORD_ALIGN(nv->logDataPageObj.dirtyByteStart);

	///this will read all the data in the data section and try to determine if we're in an unclean state.
	readDataSection();

	if(memory_section_count)
	{
		///we need to salloc memory for the memory_sections array
		nv->memory_sections = (NONVOLATILE_MEMORY_SECTION *)salloc(sizeof(NONVOLATILE_MEMORY_SECTION) *
							  memory_section_count);
		///now we must initialize every memory section
		for(i = 0; i < memory_section_count; i++)
		{
			///for each memory section, we initialise its ID, size, and max count.
			nv->memory_sections[i].memory_chunk_id  = memory_chunk_ids[i];
			nv->memory_sections[i].memory_chunk_size = memory_chunk_sizes[i];
			WORD_ALIGN(nv->memory_sections[i].memory_chunk_size);
			nv->memory_sections[i].memory_chunk_max_count = memory_chunk_max_count[i];
			nv->memory_sections[i].current_memory_pointer_offset = 0;

			if(currentMemoryLocation + (memory_chunk_sizes[i] * memory_chunk_max_count[i]) > 1024)
			{
				///it's too large to fit into our NV data, throw error here
				///todo implement error
				while(1);
			}
			///we set the start location according to the current file page start, and offset calculated by the sizes of the files we will be storing
			nv->memory_sections[i].memory_section_start_location = nv->currentFilePageStart +
					currentMemoryLocation ;
			WORD_ALIGN(nv->memory_sections[i].memory_section_start_location);
			///we increment the currentMemoryLocation according to the size of the files we wish to store here
			currentMemoryLocation += (memory_chunk_sizes[i] * memory_chunk_max_count[i]);
			///word align for good measure
			WORD_ALIGN(currentMemoryLocation);
		}
	}

	///now we calculate the latest entry's location
	readFileSection();

#ifdef BLE_STACK_SUPPORT_REQD

#else

#endif

}

const uint8_t *getNVMemorySection(MACHINE_CONFIG_ID memory_section_id)
{
	uint8_t i;

	switch(memory_section_id)
	{
		case NV_CONFIG_FILE:
			return (uint8_t *)configFile;
		case NV_LOG_FILE:
			return nv->logDataPageObj.dataStart + (nv->logDataPageObj.dataBlockSize *
												   nv->logDataPageObj.currentDataBlock);
		case NV_DIRTY_BYTE:
			if(nv->logDataPageObj.dirtyByteStart[nv->logDataPageObj.currentDirtyByte] == DIRTY_BYTE_VALUE)
			{
				nv->logDataPageObj.dirtyByteReturn = TRUE;
			}
			else
			{
				nv->logDataPageObj.dirtyByteReturn = FALSE;
			}
			return &nv->logDataPageObj.dirtyByteReturn;
		case NV_LOG_PAGE_WRITES:
			///we cast to a page header
			return (uint8_t *) & ((NV_PAGE_HEADER *)nv->currentLogPageStart)->flashPageWrittenAmount;
			break;
		case NV_FILE_PAGE_WRITES:
			///we cast to a page header
			return (uint8_t *) & ((NV_PAGE_HEADER *)nv->currentFilePageStart)->flashPageWrittenAmount;
			break;
		default:
			for(i = 0; i < nv->memory_section_count; i++)
			{
				if(nv->memory_sections[i].memory_chunk_id == memory_section_id)
				{
					///we found it, I repeat, we found it. name that reference
					return nv->memory_sections[i].memory_section_start_location +
						   nv->memory_sections[i].current_memory_pointer_offset;
				}
			}
			break;
	}

	return NULL;
}

BOOL writeNVMemorySection(MACHINE_CONFIG_ID memory_section_id, uint8_t *data, uint16_t offset,
						  uint16_t size)
{
	const uint8_t *memoryLocation;
	uint8_t i;
	switch(memory_section_id)
	{
		case NV_LOG_FILE:
			if(size > nv->logDataPageObj.dataBlockSize)
			{
				///the size we wish to store is different that what we should store
				return FALSE;
			}
			///test to see if we have reached the end of the section
			if(nv->logDataPageObj.currentDataBlock + 1 >= nv->logDataPageObj.dataSectionCount)
			{
				///we are at the end of the memory. Use the blank page next
				if(nv->uncleanPageStart)
				{
					if(nv->uncleanPageStart[0] == FLASH_LOG_PAGE_ID)
					{
						if((uint32_t)nv->uncleanPageStart >> 10 == (uint32_t)nv->logDataPageObj.dataStart >> 10)
						{
							fixUncleanPage();
						}
						else
						{
							///the unclean page can be used;
							nv->logDataPageObj.dataStart = nv->uncleanPageStart + ((uint32_t)nv->logDataPageObj.dataStart &
														   (FLASH_PAGE_SIZE - 1));
							nv->logDataPageObj.currentDataBlock = 0;
						}
					}
					else
					{
						fixUncleanPage();
					}
				}
				///ensure that we have a blank file to write to
				if(nv->blankFilePageStart)
				{
					NV_PAGE_HEADER fileHeader;
					///now we determine
					while(isNVBusy())
					{
						///wait until we aren't busy any longer
					}
					memcpy(&fileHeader, nv->currentLogPageStart, sizeof(NV_PAGE_HEADER));
					fileHeader.flashPageWrittenAmount += 1;
					//fileHeader = *(uint32_t *)nv->currentLogPageStart;
					//fileHeader += 0x00010000; ///we add 1 to the total Times written

					///ensure we don't have a null pointer
					flash_write(nv->blankFilePageStart, &fileHeader, sizeof(NV_PAGE_HEADER));

					nv->uncleanPageStart = nv->blankFilePageStart;
					nv->blankFilePageStart = 0;
					nv->logDataPageObj.dataStart = ((uint32_t)nv->logDataPageObj.dataStart &
													(FLASH_PAGE_SIZE - 1)) + nv->uncleanPageStart;
					nv->logDataPageObj.currentDataBlock = 0;
				}
				else
				{
					///we shouldn't really ever get here. Our memory is messed up
				}
			}
			else
			{
				for(i = 0; i < nv->logDataPageObj.dataBlockSize; i++)
				{
					///ensure each one
					if(*((uint8_t *)(nv->logDataPageObj.dataStart + (nv->logDataPageObj.dataBlockSize *
									 nv->logDataPageObj.currentDataBlock) + i)) == 0xFF)
					{

					}
					else
					{
						///we can just increment if we're not already on a blank part
						nv->logDataPageObj.currentDataBlock++;
						break;
					}
				}/// if the for loop exits without any else statement being called, then we won't increment(we're on a blank location)
			}
			memoryLocation = nv->logDataPageObj.dataStart + (nv->logDataPageObj.dataBlockSize *
							 nv->logDataPageObj.currentDataBlock);
			flash_write(memoryLocation, data, size);
			return TRUE;
		case NV_CONFIG_FILE:
			if(size & 0x3)
			{
				///the size isn't at least word aligned
				return FALSE;
			}
			if(size + offset > CONFIG_FILE_SIZE)
			{
				///it's too big for the config file
				return FALSE;
			}

			memoryLocation = (const uint8_t *)&configFile[offset];
			flash_write(memoryLocation, data, size);
			return TRUE;
		case NV_DIRTY_BYTE:
			if(size > 1)
			{
				///we should only have 1 byte size
				return FALSE;
			}

			if(*(data) == DIRTY_BYTE_VALUE)
			{
				if(nv->logDataPageObj.currentDirtyByte + 1 >= nv->logDataPageObj.dirtySectionCount)
				{
					///we are at the end of the memory. Use the blank page next
					if(nv->uncleanPageStart)
					{
						if(nv->uncleanPageStart[0] == FLASH_LOG_PAGE_ID)
						{
							if((uint32_t)nv->uncleanPageStart >> 10 == (uint32_t)nv->logDataPageObj.dirtyByteStart >> 10)
							{
								///we've ran out of memory, even on the rollover page
								fixUncleanPage();
							}
							else
							{
								///the unclean page can be used;
								nv->logDataPageObj.dirtyByteStart = nv->uncleanPageStart + ((uint32_t)
																	nv->logDataPageObj.dirtyByteStart & (FLASH_PAGE_SIZE - 1));
								nv->logDataPageObj.currentDirtyByte = 0;
							}
						}
						else
						{
							fixUncleanPage();
						}
					}
					///ensure that we have a blank file to write to
					if(nv->blankFilePageStart)
					{
						NV_PAGE_HEADER fileHeader;
						///now we determine
						while(isNVBusy())
						{
							///wait until we aren't busy any longer
						}
						memcpy(&fileHeader, nv->currentLogPageStart, sizeof(NV_PAGE_HEADER));
						fileHeader.flashPageWrittenAmount += 1;
						//fileHeader = *(uint32_t *)nv->currentLogPageStart;
						//fileHeader += 0x00010000; ///we add 1 to the total Times written

						///ensure we don't have a null pointer
						flash_write(nv->blankFilePageStart, &fileHeader, sizeof(NV_PAGE_HEADER));

						nv->uncleanPageStart = nv->blankFilePageStart;
						nv->blankFilePageStart = 0;
						nv->logDataPageObj.dirtyByteStart = ((uint32_t)nv->logDataPageObj.dirtyByteStart &
															 (FLASH_PAGE_SIZE - 1)) + nv->uncleanPageStart;
						nv->logDataPageObj.currentDirtyByte = 0;
					}
					else
					{
						///we shouldn't really ever get here. Our memory is messed up
					}
				}
				else
				{
					if(*(nv->logDataPageObj.dirtyByteStart + nv->logDataPageObj.currentDirtyByte) == 0xFF)
					{

					}
					else
					{
						///we can just increment if we're not already on a blank part
						nv->logDataPageObj.currentDirtyByte++;
					}
				}

			}
			else if(*(data) == CLEAN_BYTE_VALUE)
			{

			}
			else
			{
				///incorrect data to be written
				return FALSE;
			}

			memoryLocation = nv->logDataPageObj.dirtyByteStart +
							 nv->logDataPageObj.currentDirtyByte; ///word align
			flash_write(memoryLocation, data, 1); ///always one byte
			return TRUE;
		default:
			///this is the case that any of the files are being written to
			for(i = 0; i < nv->memory_section_count; i++)
			{

				if(nv->memory_sections[i].memory_chunk_id == memory_section_id)
				{
					if(offset)
					{
						///we don't support offset in this case
						return FALSE;
					}
					///we do stuff on this memory section
					///check to see if we have to roll into another page in memory
					if(nv->memory_sections[i].current_memory_chunk_counter + 1 >=
							nv->memory_sections[i].memory_chunk_max_count)
					{
						///we have reached the end of the memory section
						///we need to move on to the next section
						if(nv->uncleanPageStart)
						{
							if(nv->uncleanPageStart[0] == FLASH_FILE_PAGE_ID)
							{
								if((uint32_t)nv->memory_sections[i].memory_section_start_location >> 10 ==
										(uint32_t)nv->uncleanPageStart >> 10)
								{
									///we have written this so many times, even the unclean page has rolled over
									fixUncleanPage();
								}
								else
								{
									///here we remap the location
									nv->memory_sections[i].memory_section_start_location = (((uint32_t)
											nv->memory_sections[i].memory_section_start_location) & (FLASH_PAGE_SIZE - 1)) +
											nv->uncleanPageStart;
									nv->memory_sections[i].current_memory_chunk_counter = 0;
									nv->memory_sections[i].current_memory_pointer_offset = 0;
								}
							}
							else
							{
								///the unclean page is another page, lets clean it up
								fixUncleanPage();
							}
						}

						///make sure that we have a blank page
						if(nv->blankFilePageStart)
						{
							uint32_t fileHeader;

							while(isNVBusy())
							{
								///wait until we aren't busy any longer
							}
							fileHeader = *(uint32_t *)nv->currentFilePageStart;
							fileHeader += 0x00010000; ///we add 1 to the total Times written

							flash_write(nv->blankFilePageStart, &fileHeader, 4);
							nv->uncleanPageStart = nv->blankFilePageStart;
							nv->blankFilePageStart = 0;
							nv->memory_sections[i].memory_section_start_location = (((uint32_t)
									nv->memory_sections[i].memory_section_start_location) & (FLASH_PAGE_SIZE - 1)) +
									nv->uncleanPageStart;
							nv->memory_sections[i].current_memory_chunk_counter = 0;
							nv->memory_sections[i].current_memory_pointer_offset = 0;

						}

					}
					else
					{
						nv->memory_sections[i].current_memory_chunk_counter++;
						nv->memory_sections[i].current_memory_pointer_offset += nv->memory_sections[i].memory_chunk_size;
					}

					memoryLocation = nv->memory_sections[i].memory_section_start_location +
									 nv->memory_sections[i].current_memory_pointer_offset;
					flash_write(memoryLocation, data, nv->memory_sections[i].memory_chunk_size); ///always one byte

					return TRUE;
				}
			}
			///we didn't find that MemoryID
			return FALSE;
	}
}

BOOL isNVBusy()
{
	return nv->busyBit;
}

void nonvolatile_callback(uint32_t sys_evt)
{
	switch(sys_evt)
	{
		case NRF_EVT_FLASH_OPERATION_SUCCESS:
			nv->busyBit = 0;
			break;
		case NRF_EVT_FLASH_OPERATION_ERROR:
			///todo implement redoing flash operation
			nv->busyBit = 0;
			break;
		default:
			break;
	}
}

/**
 * STATIC FUNCTIONS
 */

static void NO_OPTIMIZATION readFileHeaders()
{
	uint8_t i;
	uint8_t uncleanPage = 0;
	const uint16_t *fileHeaderReader = (uint16_t *)fileData;

	nv->uncleanPageStart = 0; ///we want this to be null  if we have clean pages

	///check to see if we have any unclean pages
	///we check the first and last pages first, so that the first time the system is initialized, it initalized the blank page to be the last page
	if(fileData[0 * FLASH_PAGE_SIZE] == fileData[2 * FLASH_PAGE_SIZE])
	{
		///we compare the write frequency to determine which one is new(the unclean)
		if(fileHeaderReader[(0 * (FLASH_PAGE_SIZE / 2)) + 1] > fileHeaderReader[(2 *
				(FLASH_PAGE_SIZE / 2)) + 1])
		{
			uncleanPage = PAGE_ONE_UNCLEAN;
		}
		else
		{
			uncleanPage = PAGE_THREE_UNCLEAN;
		}
	}
	else if(fileData[2 * FLASH_PAGE_SIZE] == fileData[1 * FLASH_PAGE_SIZE])
	{
		///we compare the write frequency to determine which one is new(the unclean)
		if(fileHeaderReader[(1 * (FLASH_PAGE_SIZE / 2)) + 1] > fileHeaderReader[(2 *
				(FLASH_PAGE_SIZE / 2)) + 1])
		{
			uncleanPage = PAGE_TWO_UNCLEAN;
		}
		else
		{
			uncleanPage = PAGE_THREE_UNCLEAN;
		}
	}
	else if(fileData[0 * FLASH_PAGE_SIZE] == fileData[1 * FLASH_PAGE_SIZE])
	{
		///we compare the write frequency to determine which one is new(the unclean)
		if(fileHeaderReader[(0 * (FLASH_PAGE_SIZE / 2)) + 1] > fileHeaderReader[(1 *
				(FLASH_PAGE_SIZE / 2)) + 1])
		{
			uncleanPage = PAGE_ONE_UNCLEAN;
		}
		else
		{
			uncleanPage = PAGE_TWO_UNCLEAN;
		}
	}


	for(i = 0; i < 3; i++)
	{
		if((uncleanPage - 1) == i)
		{
			//if it's an unclean page set it so
			nv->uncleanPageStart = (const uint8_t *)(fileData + (i * FLASH_PAGE_SIZE));
			///there isn't any blank page
			nv->blankFilePageStart = 0;
		}
		else
		{
			///the current number isn't an unclean page.
			///assign the current page address to the respective variable
			switch(fileData[i * FLASH_PAGE_SIZE])
			{
				case FLASH_FILE_PAGE_ID:
					nv->currentFilePageStart = (const uint8_t *)(fileData + (i * FLASH_PAGE_SIZE));
					break;
				case FLASH_LOG_PAGE_ID:
					nv->currentLogPageStart = (const uint8_t *)(fileData + (i * FLASH_PAGE_SIZE));
					break;
				case 0:
					///all the pages should be blank at this point
					///ensure that all the pages are blank

					nv->currentFilePageStart = (const uint8_t *)(fileData + (0 * FLASH_PAGE_SIZE));
					nv->currentLogPageStart = (const uint8_t *)(fileData + (1 * FLASH_PAGE_SIZE));
					nv->blankFilePageStart = (const uint8_t *)(fileData + (2 * FLASH_PAGE_SIZE));

					flash_page_erase(nv->currentFilePageStart);
					flash_page_erase(nv->currentLogPageStart);
					flash_page_erase(nv->blankFilePageStart);
					if(i == 0)
					{
						NV_PAGE_HEADER newHeader;
						newHeader.flashPageWrittenAmount = 1;
						///write zeros to the padding
						memset(&newHeader.reservedPadding, 0, sizeof(NV_PAGE_HEADER) - 4);
						newHeader.reserved = 0;

						///write the flash page ID
						newHeader.pageId = FLASH_FILE_PAGE_ID;
						flash_write(nv->currentFilePageStart, (uint8_t *)&newHeader,
									sizeof(NV_PAGE_HEADER)); ///write the 16 byte header in.

						///write the log page ID
						newHeader.pageId = FLASH_LOG_PAGE_ID;
						flash_write(nv->currentLogPageStart, (uint8_t *)&newHeader, sizeof(NV_PAGE_HEADER));

						i = 4; ///break the loop
					}
					else
					{
						///todo we really don't know what to do here. our memory is completely messed up in this case
					}
					break;
				case FLASH_BLANK_PAGE_ID:
					if(uncleanPage)
					{
						///we shouldn't ever have a blank page ID if we have unclean Pages.
						///thus we initialize the memory


					}
					else
					{
						///just a blank page
						nv->blankFilePageStart = (const uint8_t *)(fileData + (i * FLASH_PAGE_SIZE));
					}
					break;
				default:
					///todo here our memory is messed up as well
					break;
			}
		}
	}
}

static BOOL readDataSection()
{
	uint16_t i;
	uint8_t last_byte;

	for(i = 1; i < nv->logDataPageObj.dirtySectionCount; i++)
	{
		///find the first empty byte
		if(nv->logDataPageObj.dirtyByteStart[i] == ERASED_UPDATE_VALUE)
		{
			///exit for loop
			break;
		}
	}
	///set the current index of the dirty byte based on the value found
	nv->logDataPageObj.currentDirtyByte = i - 1;

	uint16_t times;
	times = (nv->logDataPageObj.dataSectionCount * nv->logDataPageObj.dataBlockSize) >> 2;
	for(i = 1; i < times; i++)
	{
		///find the first empty word
		if(((uint32_t *)(nv->logDataPageObj.dataStart))[i] == BLOCK_ERASED)
		{

			///exit for loop
			break;
		}
	}
	///set the current index of the data block based on the value that we found
	nv->logDataPageObj.currentDataBlock = ((i - 1) << 2) / nv->logDataPageObj.dataBlockSize;


	if(nv->uncleanPageStart)
	{
		if(nv->uncleanPageStart[0] == FLASH_LOG_PAGE_ID)
		{
			fixUncleanPage();
		}
	}


	memcpy(&last_byte, &nv->logDataPageObj.dirtyByteStart[nv->logDataPageObj.currentDirtyByte], 1);

	if(last_byte == DIRTY_BYTE_VALUE)
	{
		nv->logDataPageObj.dirtyByteReturn = TRUE;
		///if Dirty return TRUE
		return TRUE;
	}
	else
	{
		nv->logDataPageObj.dirtyByteReturn = FALSE;
		return FALSE;
	}
}

static void readFileSection()
{
	uint8_t i;
	uint8_t j;
	uint32_t reader;

	for(i = 0; i < nv->memory_section_count; i++)
	{
		for(j = 1; j < nv->memory_sections[i].memory_chunk_max_count; j++)
		{
			reader = *((uint32_t *)(nv->memory_sections[i].memory_section_start_location +
									(j * nv->memory_sections[i].memory_chunk_size)));
			if(reader == BLOCK_ERASED)
			{
				///this is the first empty place

				break;
			}
		}
		nv->memory_sections[i].current_memory_chunk_counter = j - 1;
		nv->memory_sections[i].current_memory_pointer_offset = (j - 1) *
				nv->memory_sections[i].memory_chunk_size;
	}

	///check to see if we have an unclean page
	if(nv->uncleanPageStart)
	{
		///check to see if it's a file page
		if(nv->uncleanPageStart[0] == FLASH_FILE_PAGE_ID)
		{
			fixUncleanPage();
		}
	}


}

static BOOL fixUncleanPage()
{
	uint8_t i;
	///check to see if we have an unclean page
	if(nv->uncleanPageStart)
	{
		///check to see if it's a file page
		if(nv->uncleanPageStart[0] == FLASH_FILE_PAGE_ID)
		{
			uint16_t offset;
			uint32_t reader;

			for(i = 0; i < nv->memory_section_count; i ++)
			{
				///
				offset = (uint32_t)nv->memory_sections[i].memory_section_start_location & (FLASH_PAGE_SIZE - 1);
				reader = *((uint32_t *)(&nv->uncleanPageStart[offset]));
				if(reader == BLOCK_ERASED)
				{
					///this is a blank area
					flash_write(nv->uncleanPageStart + offset,
								&nv->memory_sections[i].memory_section_start_location[nv->memory_sections[i].current_memory_chunk_counter
										* nv->memory_sections[i].memory_chunk_size], nv->memory_sections[i].memory_chunk_size);
				}
				///change the location to the new place
				nv->memory_sections[i].memory_section_start_location = nv->uncleanPageStart + offset;
				///we reset the index to zero
				nv->memory_sections[i].current_memory_chunk_counter = 0;
				nv->memory_sections[i].current_memory_pointer_offset = 0;
			}

			nv->blankFilePageStart = nv->currentFilePageStart;
			nv->currentFilePageStart = nv->uncleanPageStart;
			///make sure that it erases the page that supposed to be blank
			flash_page_erase(nv->blankFilePageStart);
			nv->uncleanPageStart = 0;
			readFileSection();
			return TRUE;
		}
		else if(nv->uncleanPageStart[0] == FLASH_LOG_PAGE_ID)
		{
			uint16_t offset;

			///check the log value to be there or not
			///this will essentially modulo 1024 to find the offset into memory that we are looking for
			///todo remove uint64_t casts for final form
			offset = (uint32_t)nv->logDataPageObj.dataStart & (FLASH_PAGE_SIZE - 1);
			if(*((uint32_t *)(&nv->uncleanPageStart[offset])) == BLOCK_ERASED)
			{
				///if this part of memory isn't written to yet, then we should write some data
				///todo figure out a more robust way to ensure that it's the last
				flash_write(&nv->uncleanPageStart[offset],
							&nv->logDataPageObj.dataStart[nv->logDataPageObj.dataBlockSize *
									nv->logDataPageObj.currentDataBlock], nv->logDataPageObj.dataBlockSize);
			}
			offset = (uint32_t)nv->logDataPageObj.dirtyByteStart & (FLASH_PAGE_SIZE - 1);
			if(*((uint32_t *)(&nv->uncleanPageStart[offset])) == BLOCK_ERASED)
			{
				uint32_t writer;
				writer = (nv->logDataPageObj.dirtyByteStart[nv->logDataPageObj.currentDirtyByte]) | 0xFFFFFF00;
				flash_write(nv->uncleanPageStart + offset, (uint8_t *)&writer, 4);
			}

			nv->blankFilePageStart = nv->currentLogPageStart;
			nv->currentLogPageStart = nv->uncleanPageStart;
			///reset the location for the log data to new page
			offset = (uint32_t)nv->logDataPageObj.dataStart & (FLASH_PAGE_SIZE - 1);
			nv->logDataPageObj.dataStart = nv->currentLogPageStart + offset;
			offset = (uint32_t)nv->logDataPageObj.dirtyByteStart & (FLASH_PAGE_SIZE - 1);
			nv->logDataPageObj.dirtyByteStart = nv->currentLogPageStart + offset;


			///make sure the blank file is actually blank after this
			flash_page_erase(nv->blankFilePageStart);
			nv->uncleanPageStart = 0;
			readDataSection();
			return TRUE;
		}
	}

	return FALSE; ///no update
}

static void NO_OPTIMIZATION flash_page_erase(const volatile void *page_address)
{
	nv->busyBit = 1;
#ifdef BLE_STACK_SUPPORT_REQD
	if(checkBLEActive())
	{
		sd_flash_page_erase((((uint32_t)page_address) >> 10));
	}
	else
	{

		nrf_nvmc_page_erase((uint32_t)(page_address));
		nonvolatile_callback(NRF_EVT_FLASH_OPERATION_SUCCESS);
	}
#else
//
//	sd_flash_page_erase((((uint32_t)page_address) >> 10));
	nrf_nvmc_page_erase((uint32_t)(page_address));
	nonvolatile_callback(NRF_EVT_FLASH_OPERATION_SUCCESS);
#endif
    while (*((uint32_t*)(page_address)) != 0xFFFFFFFF)
    {
        // Spin while flash erase finishes
        __NOP();
    }

	nonvolatile_callback(NRF_EVT_FLASH_OPERATION_SUCCESS);
    return;
}

/**
 * this will simulate a flash write function implemented by our microcontroller
 * @param address
 * @param value
 * @param lengthInWords how many 4 byte words we wish to write into memory
 */
static void nvmc_flash_write(const uint32_t *address, const uint32_t *value, uint16_t lengthInWords)
{

#ifdef BLE_STACK_SUPPORT_REQD

	if(checkBLEActive())
	{
		sd_flash_write((uint32_t *)(address), value, lengthInWords);
	}
	else
	{
		nrf_nvmc_write_words((uint32_t)address, value, lengthInWords);
		nonvolatile_callback(NRF_EVT_FLASH_OPERATION_SUCCESS);
	}


#else
////
//	sd_flash_write((uint32_t *)(address), value, lengthInWords);
	nrf_nvmc_write_words((uint32_t)address, value, lengthInWords);
	nonvolatile_callback(NRF_EVT_FLASH_OPERATION_SUCCESS);
#endif

}

static void NO_OPTIMIZATION flash_write(const void *address, const void *value, uint16_t length)
{
	uint16_t size;
	const void *alignedAddress;
	while(nv->busyBit)
	{
		///while we're busy do nothing
	}
	nv->busyBit = 1;
	if((length & 3) || ((uint32_t)address & 3))
	{
		///we do this to simulate a 4 byte alignment requirement
		size = ((uint32_t)address & 3) + length;
		alignedAddress = (void *)((uint32_t)address & (~(uint32_t)0x3));

		WORD_ALIGN(size);
		if(size > MAX_NV_RAM_BUFFER)
		{
			///todo handle this error
		}
		memset(nv->NV_RAM_BUFFER, 0xFF, size); ///clear out the bits of memory we don't need

		memcpy(nv->NV_RAM_BUFFER + ((uint32_t)address & 3), value, length);
		///copy the ram buffer into memory, note that when actually implemented into a system, a 0xff doesn't overwrite any data
		nvmc_flash_write((const uint32_t *)alignedAddress, (uint32_t *)nv->NV_RAM_BUFFER, size >> 2);
	}
	else
	{
		nvmc_flash_write((const uint32_t *)address, (const uint32_t *)value, length >> 2);
	}

	///we should simulate the busy bit being done once asynchronous flash write is done
	nv->busyBit = 0;
}
